from io import StringIO
import logging
from molecule import util
import os


class MoleculeLog(object):

    def __init__(self):
        self._log_stream = StringIO()
        handler = logging.StreamHandler(self._log_stream)
        self._logger = logging.getLogger('takeltest')
        self._logger.handlers = []
        self._logger.addHandler(handler)
        self._logger.setLevel(logging.DEBUG)

    def debug(self, msg):
        self._logger.debug(msg)

    def get_log(self):
        return self._log_stream.getvalue()

    def print_debug(self):
        takeltest_env = \
            {k: v for (k, v) in os.environ.items() if 'TESTVARS_' in k}
        if takeltest_env:
            print('\n')
            util.print_debug('TESTVARS ENVIRONMENT',
                             util.safe_dump(takeltest_env))
        log = self.get_log()
        if log:
            if not takeltest_env:
                print('\n')
            util.print_debug('TESTVARS LOG', log)



import os
from pathlib import Path
import pytest
from takeltest.ansibleres import AnsibleLoader
from takeltest.ansibleres import AnsibleInventory
from takeltest.ansibleres import AnsibleVarsManager
from takeltest.ansibleres import AnsibleHost
from takeltest.moleculeenv import MoleculeEnv
from takeltest.moleculelog import MoleculeLog
from takeltest.moleculeplay import MoleculePlay
from takeltest.moleculebook import MoleculeBook
from takeltest.templates import Templates
from takeltest.jsonvars import JsonVars
from takeltest.pathlist import PathList
from takeltest.testpass import TestPass
from takeltest.testvars import TestVars


###########################################################
# global cleanup function
###########################################################


@pytest.fixture(scope="session", autouse=True)
def cleanup(request, moleculelog):
    if request.config.getoption("--debug"):
        request.addfinalizer(moleculelog.print_debug)


###########################################################
# command line options: testvars group
###########################################################


def pytest_addoption(parser):
    parser.addini('log', 'test')
    testvars_optiongroup = parser.getgroup("testvars")
    testvars_optiongroup.addoption(
                     "--testvars-debug-jsonvars",
                     action="store_true",
                     default=False,
                     help="print jsonvars debug information")
    testvars_optiongroup.addoption(
                     "--testvars-no-gather-facts",
                     action="store_false",
                     default=True,
                     help="do not gather ansible_facts")
    testvars_optiongroup.addoption(
                     "--testvars-no-gather-molecule",
                     action="store_false",
                     default=True,
                     help="do not resolve molecule vars")
    testvars_optiongroup.addoption(
                     "--testvars-no-gather-roles",
                     action="store_false",
                     default=True,
                     help="do not gather vars from roles")
    testvars_optiongroup.addoption(
                     "--testvars-no-extra-vars",
                     action="store_false",
                     default=True,
                     help="do not include extra vars")


###########################################################
# fixtures: testvars option booleans
###########################################################


@pytest.fixture(scope='session')
def debug_jsonvars(request):
    '''testvars option --testvars-debug-jsonvars'''
    return request.config.getoption("--testvars-debug-jsonvars")


@pytest.fixture(scope='session')
def gather_facts(request):
    '''testvars option --testvars-no-gather-facts'''
    return request.config.getoption("--testvars-no-gather-facts")


@pytest.fixture(scope='session')
def gather_molecule(request):
    '''testvars option --testvars-no-gather-molecule'''
    return request.config.getoption("--testvars-no-gather-molecule")


@pytest.fixture(scope='session')
def gather_roles(request):
    '''testvars option --testvars-no-gather-roles'''
    return request.config.getoption("--testvars-no-gather-roles")


@pytest.fixture(scope='session')
def extra_vars(request):
    '''testvars option --testvars-no-extra-vars'''
    return request.config.getoption("--testvars-no-extra-vars")


###########################################################
# fixtures: testvars option lists
###########################################################


@pytest.fixture(scope='session')
def testvars_roles_blacklist(molecule_scenario_directory):
    '''environment variable TESTVARS_ROLES_BLACKLIST'''
    try:
        blacklist = os.environ['TESTVARS_ROLES_BLACKLIST']
    except KeyError:
        return list()
    return blacklist.split(':')


@pytest.fixture(scope='session')
def testvars_roles_whitelist(molecule_scenario_directory):
    '''environment variable TESTVARS_ROLES_WHITELIST'''
    try:
        whitelist = os.environ['TESTVARS_ROLES_WHITELIST']
    except KeyError:
        return list()
    return whitelist.split(':')


@pytest.fixture(scope='session')
def testvars_extra_vars(molecule_scenario_directory):
    '''environment variable TESTVARS_EXTRA_VARS'''
    try:
        extra_vars = Path(os.environ['TESTVARS_EXTRA_VARS'])
    except KeyError:
        return list()
    return PathList(extra_vars,
                    molecule_scenario_directory).get()


###########################################################
# fixtures: ansible resources
###########################################################


@pytest.fixture(scope='session')
def ansibleloader():
    return AnsibleLoader().get()


@pytest.fixture(scope='session')
def ansibleinventory(ansibleloader,
                     inventory_file):
    return AnsibleInventory(ansibleloader,
                            inventory_file).get()


@pytest.fixture(scope='session')
def ansiblevarsmanager(ansibleloader,
                       ansibleinventory):
    return AnsibleVarsManager(ansibleloader,
                              ansibleinventory).get()


@pytest.fixture(scope='session')
def ansiblehost(ansibleinventory):
    return AnsibleHost(ansibleinventory).get()


###########################################################
# fixtures: molecule resources
###########################################################


@pytest.fixture(scope='session')
def moleculelog():
    return MoleculeLog()


@pytest.fixture(scope='session')
def molecule_ephemeral_directory(tmp_path_factory):
    '''environment variable MOLECULE_EPHEMERAL_DIRECTORY'''
    try:
        dir = Path(os.environ['MOLECULE_EPHEMERAL_DIRECTORY'])
    except KeyError:
        dir = tmp_path_factory.mktemp('molecule_ephemeral_directory')
    return dir


@pytest.fixture(scope='session')
def molecule_scenario_directory(tmp_path_factory):
    '''environment variable MOLECULE_SCENARIO_DIRECTORY'''
    try:
        dir = Path(os.environ['MOLECULE_SCENARIO_DIRECTORY'])
    except KeyError:
        dir = tmp_path_factory.mktemp('molecule_scenario_directory')
    return dir


@pytest.fixture(scope='session')
def inventory_file(molecule_ephemeral_directory):
    '''Molecule managed ansible inventory file.'''
    inventory_file = molecule_ephemeral_directory / \
        'inventory/ansible_inventory.yml'
    inventory_dir = molecule_ephemeral_directory / 'inventory'
    inventory_dir.mkdir(exist_ok=True)
    if not inventory_file.is_file():
        inventory = "localhost"
        inventory_file.write_text(inventory)
    return inventory_file


@pytest.fixture(scope='session')
def moleculeenv(moleculelog,
                molecule_ephemeral_directory,
                molecule_scenario_directory,
                gather_roles,
                testvars_roles_blacklist,
                testvars_roles_whitelist):
    return MoleculeEnv(moleculelog,
                       molecule_ephemeral_directory,
                       molecule_scenario_directory,
                       gather_roles,
                       testvars_roles_blacklist,
                       testvars_roles_whitelist)


###########################################################
# fixtures: ansible python api
###########################################################


@pytest.fixture(scope='session')
def moleculeplay(ansibleloader,
                 ansibleinventory,
                 ansiblevarsmanager,
                 ansiblehost,
                 moleculeenv):
    '''Expose ansible python api to run playbooks against a molecule host.'''
    return MoleculePlay(ansibleloader,
                        ansibleinventory,
                        ansiblevarsmanager,
                        ansiblehost,
                        moleculeenv)


@pytest.fixture(scope='session')
def moleculebook(testvars_extra_vars,
                 moleculeplay):
    '''Run an ansible playbook against a molecule host.'''
    return MoleculeBook(testvars_extra_vars,
                        moleculeplay)


###########################################################
# fixtures: testvars helpers
###########################################################


@pytest.fixture(scope='session')
def templates(moleculebook):
    return Templates(moleculebook)


@pytest.fixture(scope='session')
def jsonvars(templates,
             gather_molecule):
    return JsonVars(templates,
                    gather_molecule)


@pytest.fixture(scope='session')
def cache_key(molecule_ephemeral_directory):
    # molecule_ephemeral_directory should be unique for each scenario
    return 'testvars' + str(molecule_ephemeral_directory)


###########################################################
# fixtures: molecule/testinfra/ansible helpers
###########################################################


@pytest.fixture(scope='session')
def testpass(moleculebook):
    '''Provide access to the ansible passwordstore lookup plugin.'''
    return TestPass(moleculebook).testpass


@pytest.fixture(scope='session')
def testvars(request,
             moleculelog,
             debug_jsonvars,
             moleculebook,
             jsonvars,
             gather_facts,
             extra_vars,
             cache_key):
    '''Expose ansible variables and facts of a molecule test scenario.'''
    testvars = TestVars.get_cache(request, cache_key)
    if testvars is None:
        testvars_object = TestVars(
                 moleculelog,
                 debug_jsonvars,
                 moleculebook,
                 jsonvars,
                 gather_facts,
                 extra_vars)
        testvars = testvars_object.get_testvars()
        testvars_object.set_cache(request, cache_key)
    return testvars



from takeltest.exceptions import MoleculeBookRunError


class MoleculeBook(object):
    '''Run an ansible playbook against a molecule host'''

    def __init__(self,
                 testvars_extra_vars,
                 moleculeplay):
        self._moleculeplay = moleculeplay
        self._testvars_extra_vars = testvars_extra_vars
        self._playbook = dict()
        self.create()

    def get(self):
        '''Get the ansible playbook'''
        return self._playbook

    def set(self, playbook):
        '''Set an ansible playbook'''
        self._playbook = playbook

    def create(self,
               gather_facts=True,
               gather_roles=True,
               extra_vars=True,
               host=None):
        '''Create an ansible playbook using the ansible python api.'''
        if host is None:
            host = self._moleculeplay.get_host()

        playbook = dict(
            name="ansible playbook",
            hosts=str(host),
            gather_facts=str(gather_facts),
            vars_files=list(),
            roles=list(),
            tasks=list(),
        )

        # include extra vars files
        if extra_vars:
            for path in self._get_extra_vars_():
                playbook['vars_files'].append(str(path))

        # include roles
        if gather_roles:
            for role in self._moleculeplay.get_roles():
                playbook['roles'].append(dict(name=role, when='False'))

        self._playbook = playbook

    def add_task_debug(self, msg):
        '''Add a task using the ansible debug module'''
        task = dict(action=dict(module='debug', args=dict(msg=msg)))
        self._playbook['tasks'].append(task)

    def add_task_include_vars_dir(self, vars_dir):
        '''Add a task using the ansible include_vars module'''
        args = dict(dir=str(vars_dir))
        task = dict(action=dict(module='include_vars', args=args))
        self._playbook['tasks'].append(task)

    def get_vars(self, gather_facts=True, extra_vars=True):
        '''Return ansible facts and vars of a molecule host.

        Args:
            gather_facts (bool): gather ansible_facts from a molecule host?
                Defaults to True.
                A playbook will be run with ``gather_facts:true``.
            extra_vars (bool): include extra vars from vars files?
                Defaults to True.
                An include_vars task will be added to include extra vars files
                specified in the environment variable TESTVARS_EXTRA_VARS

        Returns:
            vars (dict): resolved ansible variables and facts
        '''
        vars = dict()

        # self.create sets gather_facts=True by default so the ansible facts
        # of the default molecule host will be in result[0]['ansible_facts']
        self.create(gather_facts=gather_facts, extra_vars=extra_vars)

        # the ansible variables will be in result[1]['msg']
        self.add_task_debug('{{ vars }}')

        result = self.run()

        if gather_facts:
            try:
                vars = result[1]['msg']
            except (IndexError, KeyError):
                raise MoleculeBookRunError(
                    'Unable to gather ansible vars and facts.')
        else:
            try:
                vars = result[0]['msg']
            except (IndexError, KeyError):
                raise MoleculeBookRunError(
                    'Unable to gather ansible vars.')

        try:
            ansible_hostname = vars['ansible_play_hosts'][0]
            hostvars = vars['hostvars'][ansible_hostname]
            del vars['hostvars']
            vars.update(hostvars)
        except (IndexError, KeyError):
            pass

        return vars

    def read_vars(self):
        '''Return ansible variables without running a playbook.'''
        self.create(gather_facts=False, host='localhost')
        vars = self._moleculeplay.read_vars(self._playbook)
        return vars

    def run(self):
        '''Run the ansible playbook'''
        return self._moleculeplay.run_playbook(self._playbook)

    def _get_extra_vars_(self):
        return self._testvars_extra_vars

    def _get_molecule_scenario_directory_(self):
        return self._moleculeplay.get_molecule_scenario_directory()



import takeltest.ansibleres
import takeltest.ansiblerun
from takeltest.boilerplate import hosts  # noqa F401
import takeltest.exceptions
import takeltest.jsonvars
import takeltest.moleculebook
import takeltest.moleculeenv
import takeltest.moleculelog
import takeltest.moleculeplay
import takeltest.plugin
import takeltest.templates
import takeltest.testpass
import takeltest.testvars  # noqa F401

name = "takeltest"
